__all__ = (
    "Success",
    "Failure",
)

import resultful

from typing import (
    Any,
    TypeVar,
    Generic,
    Literal,
    Final,
)

ValueType = TypeVar("ValueType")
ErrorType = TypeVar("ErrorType")


class Success(Generic[ValueType]):

    __slots__ = ("value",)

    def __init__(self, value: ValueType) -> None:
        self.value: Final[ValueType] = value

    def __repr__(self) -> str:
        return f"{resultful.__name__}.{type(self).__name__}({self.value!r})"

    def __bool__(self) -> Literal[True]:
        return True

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented

        return self.value == other.value

    @property
    def is_success(self) -> Literal[True]:
        return True

    @property
    def is_failure(self) -> Literal[False]:
        return False


class Failure(Generic[ErrorType]):

    __slots__ = ("error",)

    def __init__(self, error: ErrorType) -> None:
        self.error: Final[ErrorType] = error

    def __repr__(self) -> str:
        return f"{resultful.__name__}.{type(self).__name__}({self.error!r})"

    def __bool__(self) -> Literal[False]:
        return False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented

        return self.error == other.error

    @property
    def is_success(self) -> Literal[False]:
        return False

    @property
    def is_failure(self) -> Literal[True]:
        return True



__all__ = (
    "unsafe",
    "success",
    "failure",
    "unwrap_success",
    "unwrap_failure",
    "Result",
    "AlwaysSuccess",
    "AlwaysFailure",
    "ResultType",
    "NoResult",
    "NoResultType",
)

from typing import (
    overload,
    TypeVar,
    Tuple,
    Union,
    Final,
)

from resultful.impl.result import (
    Success,
    Failure,
)

from resultful.impl.no_result import (
    NoResult,
    NoResultType,
)

ValueType = TypeVar("ValueType")
ErrorType = TypeVar("ErrorType", bound=BaseException)

Result = Union[Success[ValueType], Failure[ErrorType]]

AlwaysSuccess = Union[Success[ValueType]]
AlwaysFailure = Union[Failure[ErrorType]]

ResultType: Final[Tuple[type, ...]] = (Success, Failure)


# noinspection PyUnresolvedReferences
def unsafe(
    result: Result[ValueType, ErrorType],
) -> ValueType:

    """
    Unwraps the value from given result in unsafe way.

    :param result: result to be unwrapped

    :raises ErrorType: When result is a :class:`Failure`.
    """

    if result:
        return result.value

    raise result.error


# TODO(kprzybyla): Remove "type: ignore" once below feature will be implemented
#                  https://github.com/python/typing/issues/599


@overload
def success(  # type: ignore
    value: Failure[ErrorType],
    /,
) -> NoResultType:
    ...


@overload
def success(
    value: Success[ValueType],
    /,
) -> Success[ValueType]:
    ...


@overload
def success(
    value: ValueType,
    /,
) -> Success[ValueType]:
    ...


def success(  # type: ignore
    value: Union[Failure[ErrorType], Success[ValueType], ValueType],
    /,
) -> Union[NoResultType, Success[ValueType]]:

    """
    Returns value wrapped in :class:`Success`.

    If value is a :class:`Failure`, returns NoResult.
    If value is a :class:`Success`, returns that :class:`Success`.

    :param value: any value
    """

    if isinstance(value, Failure):
        return NoResult

    if isinstance(value, Success):
        return value

    return Success(value)


@overload
def failure(
    error: Success[ValueType],
    /,
) -> NoResultType:
    ...


@overload
def failure(
    error: Failure[ErrorType],
    /,
) -> Failure[ErrorType]:
    ...


@overload
def failure(
    error: ErrorType,
    /,
) -> Failure[ErrorType]:
    ...


def failure(
    error: Union[Success[ValueType], Failure[ErrorType], ErrorType],
    /,
) -> Union[NoResultType, Failure[ErrorType]]:

    """
    Returns error wrapped in :class:`Failure`.

    If error is a :class:`Success`, returns NoResult.
    If error is a :class:`Failure`, returns that :class:`Failure`.

    :param error: any exception
    """

    if isinstance(error, Success):
        return NoResult

    if isinstance(error, Failure):
        return error

    return Failure(error)


@overload
def unwrap_success(
    result: Failure[ErrorType],
) -> NoResultType:
    ...


@overload
def unwrap_success(
    result: Success[ValueType],
) -> ValueType:
    ...


# noinspection PyUnresolvedReferences
def unwrap_success(
    result: Result[ValueType, ErrorType],
) -> Union[NoResultType, ValueType]:

    """
    Unwraps the value from given :class:`Success`.

    :param result: :class:`Success`
    """

    if isinstance(result, Failure):
        return NoResult

    return result.value


@overload
def unwrap_failure(
    result: Success[ValueType],
) -> NoResultType:
    ...


@overload
def unwrap_failure(
    result: Failure[ErrorType],
) -> ErrorType:
    ...


# noinspection PyUnresolvedReferences
def unwrap_failure(
    result: Result[ValueType, ErrorType],
) -> Union[NoResultType, ErrorType]:

    """
    Unwraps the error from given :class:`Failure`.

    :param result: :class:`Failure`
    """

    if isinstance(result, Success):
        return NoResult

    return result.error



project = "resultful"
author = "Krzysztof Przybyła"

language = "en"
master_doc = "index"
source_suffix = ".rst"
html_theme = "sphinx_rtd_theme"
autodoc_member_order = "bysource"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]



from setuptools import (
    setup,
)


def main():
    setup(
        python_requires=">= 3.8",
        setup_requires=[
            "setuptools",
            "setuptools_scm[toml]",
        ],
        extras_require={
            "format": [
                "black == 20.8b1",
                "flake8 ~= 3.8.0",
            ],
            "test": [
                "mypy ~= 0.782",
                "pytest ~= 6.0.1",
                "pytest-cov ~= 2.8.0",
                "hypothesis ~= 5.6.0",
            ],
            "docs": [
                "sphinx ~= 3.0.0",
                "sphinx_rtd_theme ~= 0.4.3",
                "sphinx_autodoc_typehints ~= 1.10.0",
            ],
            "deploy": [
                "wheel",
                "twine",
            ],
        },
    )


if __name__ == "__main__":
    main()



__all__ = (
    "Result",
    "AlwaysSuccess",
    "AlwaysFailure",
    "ResultType",
    "NoResultType",
    "unsafe",
    "success",
    "failure",
    "unwrap_success",
    "unwrap_failure",
    "NoResult",
)

# Annotations

from .result import (
    Result,
    AlwaysSuccess,
    AlwaysFailure,
)

# Types

from .result import (
    ResultType,
    NoResultType,
)

# Concretes

from .result import (
    unsafe,
    success,
    failure,
    unwrap_success,
    unwrap_failure,
    NoResult,
)



__all__ = (
    "NoResult",
    "NoResultType",
)

import enum
import resultful

from typing import (
    Literal,
    Final,
)


class NoResultType(enum.Enum):

    NO_RESULT = enum.auto()

    def __repr__(self) -> str:
        return f"{resultful.__name__}.NoResult"

    # noinspection PyTypeChecker
    def __bool__(self) -> Literal[False]:
        return False


NoResult: Final[Literal[NoResultType.NO_RESULT]] = NoResultType.NO_RESULT
